# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: yoti.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="yoti.proto",
    package="aea.fetchai.yoti.v0_1_0",
    syntax="proto3",
    serialized_options=None,
    serialized_pb=b'\n\nyoti.proto\x12\x17\x61\x65\x61.fetchai.yoti.v0_1_0"\xaf\x04\n\x0bYotiMessage\x12H\n\x05\x65rror\x18\x05 \x01(\x0b\x32\x37.aea.fetchai.yoti.v0_1_0.YotiMessage.Error_PerformativeH\x00\x12T\n\x0bget_profile\x18\x06 \x01(\x0b\x32=.aea.fetchai.yoti.v0_1_0.YotiMessage.Get_Profile_PerformativeH\x00\x12L\n\x07profile\x18\x07 \x01(\x0b\x32\x39.aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_PerformativeH\x00\x1aL\n\x18Get_Profile_Performative\x12\r\n\x05token\x18\x01 \x01(\t\x12\x13\n\x0b\x64otted_path\x18\x02 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x03 \x03(\t\x1a\x96\x01\n\x14Profile_Performative\x12Q\n\x04info\x18\x01 \x03(\x0b\x32\x43.aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative.InfoEntry\x1a+\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a;\n\x12\x45rror_Performative\x12\x12\n\nerror_code\x18\x01 \x01(\x05\x12\x11\n\terror_msg\x18\x02 \x01(\tB\x0e\n\x0cperformativeb\x06proto3',
)


_YOTIMESSAGE_GET_PROFILE_PERFORMATIVE = _descriptor.Descriptor(
    name="Get_Profile_Performative",
    full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Get_Profile_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="token",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Get_Profile_Performative.token",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="dotted_path",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Get_Profile_Performative.dotted_path",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="args",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Get_Profile_Performative.args",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=293,
    serialized_end=369,
)

_YOTIMESSAGE_PROFILE_PERFORMATIVE_INFOENTRY = _descriptor.Descriptor(
    name="InfoEntry",
    full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative.InfoEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative.InfoEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative.InfoEntry.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=479,
    serialized_end=522,
)

_YOTIMESSAGE_PROFILE_PERFORMATIVE = _descriptor.Descriptor(
    name="Profile_Performative",
    full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="info",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative.info",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[_YOTIMESSAGE_PROFILE_PERFORMATIVE_INFOENTRY,],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=372,
    serialized_end=522,
)

_YOTIMESSAGE_ERROR_PERFORMATIVE = _descriptor.Descriptor(
    name="Error_Performative",
    full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Error_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="error_code",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Error_Performative.error_code",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="error_msg",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.Error_Performative.error_msg",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=524,
    serialized_end=583,
)

_YOTIMESSAGE = _descriptor.Descriptor(
    name="YotiMessage",
    full_name="aea.fetchai.yoti.v0_1_0.YotiMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="error",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.error",
            index=0,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="get_profile",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.get_profile",
            index=1,
            number=6,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="profile",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.profile",
            index=2,
            number=7,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[
        _YOTIMESSAGE_GET_PROFILE_PERFORMATIVE,
        _YOTIMESSAGE_PROFILE_PERFORMATIVE,
        _YOTIMESSAGE_ERROR_PERFORMATIVE,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="performative",
            full_name="aea.fetchai.yoti.v0_1_0.YotiMessage.performative",
            index=0,
            containing_type=None,
            fields=[],
        ),
    ],
    serialized_start=40,
    serialized_end=599,
)

_YOTIMESSAGE_GET_PROFILE_PERFORMATIVE.containing_type = _YOTIMESSAGE
_YOTIMESSAGE_PROFILE_PERFORMATIVE_INFOENTRY.containing_type = (
    _YOTIMESSAGE_PROFILE_PERFORMATIVE
)
_YOTIMESSAGE_PROFILE_PERFORMATIVE.fields_by_name[
    "info"
].message_type = _YOTIMESSAGE_PROFILE_PERFORMATIVE_INFOENTRY
_YOTIMESSAGE_PROFILE_PERFORMATIVE.containing_type = _YOTIMESSAGE
_YOTIMESSAGE_ERROR_PERFORMATIVE.containing_type = _YOTIMESSAGE
_YOTIMESSAGE.fields_by_name["error"].message_type = _YOTIMESSAGE_ERROR_PERFORMATIVE
_YOTIMESSAGE.fields_by_name[
    "get_profile"
].message_type = _YOTIMESSAGE_GET_PROFILE_PERFORMATIVE
_YOTIMESSAGE.fields_by_name["profile"].message_type = _YOTIMESSAGE_PROFILE_PERFORMATIVE
_YOTIMESSAGE.oneofs_by_name["performative"].fields.append(
    _YOTIMESSAGE.fields_by_name["error"]
)
_YOTIMESSAGE.fields_by_name["error"].containing_oneof = _YOTIMESSAGE.oneofs_by_name[
    "performative"
]
_YOTIMESSAGE.oneofs_by_name["performative"].fields.append(
    _YOTIMESSAGE.fields_by_name["get_profile"]
)
_YOTIMESSAGE.fields_by_name[
    "get_profile"
].containing_oneof = _YOTIMESSAGE.oneofs_by_name["performative"]
_YOTIMESSAGE.oneofs_by_name["performative"].fields.append(
    _YOTIMESSAGE.fields_by_name["profile"]
)
_YOTIMESSAGE.fields_by_name["profile"].containing_oneof = _YOTIMESSAGE.oneofs_by_name[
    "performative"
]
DESCRIPTOR.message_types_by_name["YotiMessage"] = _YOTIMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

YotiMessage = _reflection.GeneratedProtocolMessageType(
    "YotiMessage",
    (_message.Message,),
    {
        "Get_Profile_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_Profile_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _YOTIMESSAGE_GET_PROFILE_PERFORMATIVE,
                "__module__": "yoti_pb2"
                # @@protoc_insertion_point(class_scope:aea.fetchai.yoti.v0_1_0.YotiMessage.Get_Profile_Performative)
            },
        ),
        "Profile_Performative": _reflection.GeneratedProtocolMessageType(
            "Profile_Performative",
            (_message.Message,),
            {
                "InfoEntry": _reflection.GeneratedProtocolMessageType(
                    "InfoEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _YOTIMESSAGE_PROFILE_PERFORMATIVE_INFOENTRY,
                        "__module__": "yoti_pb2"
                        # @@protoc_insertion_point(class_scope:aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative.InfoEntry)
                    },
                ),
                "DESCRIPTOR": _YOTIMESSAGE_PROFILE_PERFORMATIVE,
                "__module__": "yoti_pb2"
                # @@protoc_insertion_point(class_scope:aea.fetchai.yoti.v0_1_0.YotiMessage.Profile_Performative)
            },
        ),
        "Error_Performative": _reflection.GeneratedProtocolMessageType(
            "Error_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _YOTIMESSAGE_ERROR_PERFORMATIVE,
                "__module__": "yoti_pb2"
                # @@protoc_insertion_point(class_scope:aea.fetchai.yoti.v0_1_0.YotiMessage.Error_Performative)
            },
        ),
        "DESCRIPTOR": _YOTIMESSAGE,
        "__module__": "yoti_pb2"
        # @@protoc_insertion_point(class_scope:aea.fetchai.yoti.v0_1_0.YotiMessage)
    },
)
_sym_db.RegisterMessage(YotiMessage)
_sym_db.RegisterMessage(YotiMessage.Get_Profile_Performative)
_sym_db.RegisterMessage(YotiMessage.Profile_Performative)
_sym_db.RegisterMessage(YotiMessage.Profile_Performative.InfoEntry)
_sym_db.RegisterMessage(YotiMessage.Error_Performative)


_YOTIMESSAGE_PROFILE_PERFORMATIVE_INFOENTRY._options = None
# @@protoc_insertion_point(module_scope)
