// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.25.0
// 	protoc        v3.11.4
// source: default.proto

package aea_fetchai_default_v0_1_0

import (
	proto "github.com/golang/protobuf/proto"
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// This is a compile-time assertion that a sufficiently up-to-date version
// of the legacy proto package is being used.
const _ = proto.ProtoPackageIsVersion4

type DefaultMessage_ErrorCode_ErrorCodeEnum int32

const (
	DefaultMessage_ErrorCode_UNSUPPORTED_PROTOCOL DefaultMessage_ErrorCode_ErrorCodeEnum = 0
	DefaultMessage_ErrorCode_DECODING_ERROR       DefaultMessage_ErrorCode_ErrorCodeEnum = 1
	DefaultMessage_ErrorCode_INVALID_MESSAGE      DefaultMessage_ErrorCode_ErrorCodeEnum = 2
	DefaultMessage_ErrorCode_UNSUPPORTED_SKILL    DefaultMessage_ErrorCode_ErrorCodeEnum = 3
	DefaultMessage_ErrorCode_INVALID_DIALOGUE     DefaultMessage_ErrorCode_ErrorCodeEnum = 4
)

// Enum value maps for DefaultMessage_ErrorCode_ErrorCodeEnum.
var (
	DefaultMessage_ErrorCode_ErrorCodeEnum_name = map[int32]string{
		0: "UNSUPPORTED_PROTOCOL",
		1: "DECODING_ERROR",
		2: "INVALID_MESSAGE",
		3: "UNSUPPORTED_SKILL",
		4: "INVALID_DIALOGUE",
	}
	DefaultMessage_ErrorCode_ErrorCodeEnum_value = map[string]int32{
		"UNSUPPORTED_PROTOCOL": 0,
		"DECODING_ERROR":       1,
		"INVALID_MESSAGE":      2,
		"UNSUPPORTED_SKILL":    3,
		"INVALID_DIALOGUE":     4,
	}
)

func (x DefaultMessage_ErrorCode_ErrorCodeEnum) Enum() *DefaultMessage_ErrorCode_ErrorCodeEnum {
	p := new(DefaultMessage_ErrorCode_ErrorCodeEnum)
	*p = x
	return p
}

func (x DefaultMessage_ErrorCode_ErrorCodeEnum) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (DefaultMessage_ErrorCode_ErrorCodeEnum) Descriptor() protoreflect.EnumDescriptor {
	return file_default_proto_enumTypes[0].Descriptor()
}

func (DefaultMessage_ErrorCode_ErrorCodeEnum) Type() protoreflect.EnumType {
	return &file_default_proto_enumTypes[0]
}

func (x DefaultMessage_ErrorCode_ErrorCodeEnum) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use DefaultMessage_ErrorCode_ErrorCodeEnum.Descriptor instead.
func (DefaultMessage_ErrorCode_ErrorCodeEnum) EnumDescriptor() ([]byte, []int) {
	return file_default_proto_rawDescGZIP(), []int{0, 0, 0}
}

type DefaultMessage struct {
	state         protoimpl.MessageState
	sizeCache     protoimpl.SizeCache
	unknownFields protoimpl.UnknownFields

	// Types that are assignable to Performative:
	//	*DefaultMessage_Bytes
	//	*DefaultMessage_End
	//	*DefaultMessage_Error
	Performative isDefaultMessage_Performative `protobuf_oneof:"performative"`
}

func (x *DefaultMessage) Reset() {
	*x = DefaultMessage{}
	if protoimpl.UnsafeEnabled {
		mi := &file_default_proto_msgTypes[0]
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		ms.StoreMessageInfo(mi)
	}
}

func (x *DefaultMessage) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DefaultMessage) ProtoMessage() {}

func (x *DefaultMessage) ProtoReflect() protoreflect.Message {
	mi := &file_default_proto_msgTypes[0]
	if protoimpl.UnsafeEnabled && x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DefaultMessage.ProtoReflect.Descriptor instead.
func (*DefaultMessage) Descriptor() ([]byte, []int) {
	return file_default_proto_rawDescGZIP(), []int{0}
}

func (m *DefaultMessage) GetPerformative() isDefaultMessage_Performative {
	if m != nil {
		return m.Performative
	}
	return nil
}

func (x *DefaultMessage) GetBytes() *DefaultMessage_Bytes_Performative {
	if x, ok := x.GetPerformative().(*DefaultMessage_Bytes); ok {
		return x.Bytes
	}
	return nil
}

func (x *DefaultMessage) GetEnd() *DefaultMessage_End_Performative {
	if x, ok := x.GetPerformative().(*DefaultMessage_End); ok {
		return x.End
	}
	return nil
}

func (x *DefaultMessage) GetError() *DefaultMessage_Error_Performative {
	if x, ok := x.GetPerformative().(*DefaultMessage_Error); ok {
		return x.Error
	}
	return nil
}

type isDefaultMessage_Performative interface {
	isDefaultMessage_Performative()
}

type DefaultMessage_Bytes struct {
	Bytes *DefaultMessage_Bytes_Performative `protobuf:"bytes,5,opt,name=bytes,proto3,oneof"`
}

type DefaultMessage_End struct {
	End *DefaultMessage_End_Performative `protobuf:"bytes,6,opt,name=end,proto3,oneof"`
}

type DefaultMessage_Error struct {
	Error *DefaultMessage_Error_Performative `protobuf:"bytes,7,opt,name=error,proto3,oneof"`
}

func (*DefaultMessage_Bytes) isDefaultMessage_Performative() {}

func (*DefaultMessage_End) isDefaultMessage_Performative() {}

func (*DefaultMessage_Error) isDefaultMessage_Performative() {}

// Custom Types
type DefaultMessage_ErrorCode struct {
	state         protoimpl.MessageState
	sizeCache     protoimpl.SizeCache
	unknownFields protoimpl.UnknownFields

	ErrorCode DefaultMessage_ErrorCode_ErrorCodeEnum `protobuf:"varint,1,opt,name=error_code,json=errorCode,proto3,enum=aea.fetchai.default.v0_1_0.DefaultMessage_ErrorCode_ErrorCodeEnum" json:"error_code,omitempty"`
}

func (x *DefaultMessage_ErrorCode) Reset() {
	*x = DefaultMessage_ErrorCode{}
	if protoimpl.UnsafeEnabled {
		mi := &file_default_proto_msgTypes[1]
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		ms.StoreMessageInfo(mi)
	}
}

func (x *DefaultMessage_ErrorCode) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DefaultMessage_ErrorCode) ProtoMessage() {}

func (x *DefaultMessage_ErrorCode) ProtoReflect() protoreflect.Message {
	mi := &file_default_proto_msgTypes[1]
	if protoimpl.UnsafeEnabled && x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DefaultMessage_ErrorCode.ProtoReflect.Descriptor instead.
func (*DefaultMessage_ErrorCode) Descriptor() ([]byte, []int) {
	return file_default_proto_rawDescGZIP(), []int{0, 0}
}

func (x *DefaultMessage_ErrorCode) GetErrorCode() DefaultMessage_ErrorCode_ErrorCodeEnum {
	if x != nil {
		return x.ErrorCode
	}
	return DefaultMessage_ErrorCode_UNSUPPORTED_PROTOCOL
}

// Performatives and contents
type DefaultMessage_Bytes_Performative struct {
	state         protoimpl.MessageState
	sizeCache     protoimpl.SizeCache
	unknownFields protoimpl.UnknownFields

	Content []byte `protobuf:"bytes,1,opt,name=content,proto3" json:"content,omitempty"`
}

func (x *DefaultMessage_Bytes_Performative) Reset() {
	*x = DefaultMessage_Bytes_Performative{}
	if protoimpl.UnsafeEnabled {
		mi := &file_default_proto_msgTypes[2]
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		ms.StoreMessageInfo(mi)
	}
}

func (x *DefaultMessage_Bytes_Performative) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DefaultMessage_Bytes_Performative) ProtoMessage() {}

func (x *DefaultMessage_Bytes_Performative) ProtoReflect() protoreflect.Message {
	mi := &file_default_proto_msgTypes[2]
	if protoimpl.UnsafeEnabled && x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DefaultMessage_Bytes_Performative.ProtoReflect.Descriptor instead.
func (*DefaultMessage_Bytes_Performative) Descriptor() ([]byte, []int) {
	return file_default_proto_rawDescGZIP(), []int{0, 1}
}

func (x *DefaultMessage_Bytes_Performative) GetContent() []byte {
	if x != nil {
		return x.Content
	}
	return nil
}

type DefaultMessage_Error_Performative struct {
	state         protoimpl.MessageState
	sizeCache     protoimpl.SizeCache
	unknownFields protoimpl.UnknownFields

	ErrorCode *DefaultMessage_ErrorCode `protobuf:"bytes,1,opt,name=error_code,json=errorCode,proto3" json:"error_code,omitempty"`
	ErrorMsg  string                    `protobuf:"bytes,2,opt,name=error_msg,json=errorMsg,proto3" json:"error_msg,omitempty"`
	ErrorData map[string][]byte         `protobuf:"bytes,3,rep,name=error_data,json=errorData,proto3" json:"error_data,omitempty" protobuf_key:"bytes,1,opt,name=key,proto3" protobuf_val:"bytes,2,opt,name=value,proto3"`
}

func (x *DefaultMessage_Error_Performative) Reset() {
	*x = DefaultMessage_Error_Performative{}
	if protoimpl.UnsafeEnabled {
		mi := &file_default_proto_msgTypes[3]
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		ms.StoreMessageInfo(mi)
	}
}

func (x *DefaultMessage_Error_Performative) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DefaultMessage_Error_Performative) ProtoMessage() {}

func (x *DefaultMessage_Error_Performative) ProtoReflect() protoreflect.Message {
	mi := &file_default_proto_msgTypes[3]
	if protoimpl.UnsafeEnabled && x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DefaultMessage_Error_Performative.ProtoReflect.Descriptor instead.
func (*DefaultMessage_Error_Performative) Descriptor() ([]byte, []int) {
	return file_default_proto_rawDescGZIP(), []int{0, 2}
}

func (x *DefaultMessage_Error_Performative) GetErrorCode() *DefaultMessage_ErrorCode {
	if x != nil {
		return x.ErrorCode
	}
	return nil
}

func (x *DefaultMessage_Error_Performative) GetErrorMsg() string {
	if x != nil {
		return x.ErrorMsg
	}
	return ""
}

func (x *DefaultMessage_Error_Performative) GetErrorData() map[string][]byte {
	if x != nil {
		return x.ErrorData
	}
	return nil
}

type DefaultMessage_End_Performative struct {
	state         protoimpl.MessageState
	sizeCache     protoimpl.SizeCache
	unknownFields protoimpl.UnknownFields
}

func (x *DefaultMessage_End_Performative) Reset() {
	*x = DefaultMessage_End_Performative{}
	if protoimpl.UnsafeEnabled {
		mi := &file_default_proto_msgTypes[4]
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		ms.StoreMessageInfo(mi)
	}
}

func (x *DefaultMessage_End_Performative) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DefaultMessage_End_Performative) ProtoMessage() {}

func (x *DefaultMessage_End_Performative) ProtoReflect() protoreflect.Message {
	mi := &file_default_proto_msgTypes[4]
	if protoimpl.UnsafeEnabled && x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DefaultMessage_End_Performative.ProtoReflect.Descriptor instead.
func (*DefaultMessage_End_Performative) Descriptor() ([]byte, []int) {
	return file_default_proto_rawDescGZIP(), []int{0, 3}
}

var File_default_proto protoreflect.FileDescriptor

var file_default_proto_rawDesc = []byte{
	0x0a, 0x0d, 0x64, 0x65, 0x66, 0x61, 0x75, 0x6c, 0x74, 0x2e, 0x70, 0x72, 0x6f, 0x74, 0x6f, 0x12,
	0x1a, 0x61, 0x65, 0x61, 0x2e, 0x66, 0x65, 0x74, 0x63, 0x68, 0x61, 0x69, 0x2e, 0x64, 0x65, 0x66,
	0x61, 0x75, 0x6c, 0x74, 0x2e, 0x76, 0x30, 0x5f, 0x31, 0x5f, 0x30, 0x22, 0x89, 0x07, 0x0a, 0x0e,
	0x44, 0x65, 0x66, 0x61, 0x75, 0x6c, 0x74, 0x4d, 0x65, 0x73, 0x73, 0x61, 0x67, 0x65, 0x12, 0x55,
	0x0a, 0x05, 0x62, 0x79, 0x74, 0x65, 0x73, 0x18, 0x05, 0x20, 0x01, 0x28, 0x0b, 0x32, 0x3d, 0x2e,
	0x61, 0x65, 0x61, 0x2e, 0x66, 0x65, 0x74, 0x63, 0x68, 0x61, 0x69, 0x2e, 0x64, 0x65, 0x66, 0x61,
	0x75, 0x6c, 0x74, 0x2e, 0x76, 0x30, 0x5f, 0x31, 0x5f, 0x30, 0x2e, 0x44, 0x65, 0x66, 0x61, 0x75,
	0x6c, 0x74, 0x4d, 0x65, 0x73, 0x73, 0x61, 0x67, 0x65, 0x2e, 0x42, 0x79, 0x74, 0x65, 0x73, 0x5f,
	0x50, 0x65, 0x72, 0x66, 0x6f, 0x72, 0x6d, 0x61, 0x74, 0x69, 0x76, 0x65, 0x48, 0x00, 0x52, 0x05,
	0x62, 0x79, 0x74, 0x65, 0x73, 0x12, 0x4f, 0x0a, 0x03, 0x65, 0x6e, 0x64, 0x18, 0x06, 0x20, 0x01,
	0x28, 0x0b, 0x32, 0x3b, 0x2e, 0x61, 0x65, 0x61, 0x2e, 0x66, 0x65, 0x74, 0x63, 0x68, 0x61, 0x69,
	0x2e, 0x64, 0x65, 0x66, 0x61, 0x75, 0x6c, 0x74, 0x2e, 0x76, 0x30, 0x5f, 0x31, 0x5f, 0x30, 0x2e,
	0x44, 0x65, 0x66, 0x61, 0x75, 0x6c, 0x74, 0x4d, 0x65, 0x73, 0x73, 0x61, 0x67, 0x65, 0x2e, 0x45,
	0x6e, 0x64, 0x5f, 0x50, 0x65, 0x72, 0x66, 0x6f, 0x72, 0x6d, 0x61, 0x74, 0x69, 0x76, 0x65, 0x48,
	0x00, 0x52, 0x03, 0x65, 0x6e, 0x64, 0x12, 0x55, 0x0a, 0x05, 0x65, 0x72, 0x72, 0x6f, 0x72, 0x18,
	0x07, 0x20, 0x01, 0x28, 0x0b, 0x32, 0x3d, 0x2e, 0x61, 0x65, 0x61, 0x2e, 0x66, 0x65, 0x74, 0x63,
	0x68, 0x61, 0x69, 0x2e, 0x64, 0x65, 0x66, 0x61, 0x75, 0x6c, 0x74, 0x2e, 0x76, 0x30, 0x5f, 0x31,
	0x5f, 0x30, 0x2e, 0x44, 0x65, 0x66, 0x61, 0x75, 0x6c, 0x74, 0x4d, 0x65, 0x73, 0x73, 0x61, 0x67,
	0x65, 0x2e, 0x45, 0x72, 0x72, 0x6f, 0x72, 0x5f, 0x50, 0x65, 0x72, 0x66, 0x6f, 0x72, 0x6d, 0x61,
	0x74, 0x69, 0x76, 0x65, 0x48, 0x00, 0x52, 0x05, 0x65, 0x72, 0x72, 0x6f, 0x72, 0x1a, 0xef, 0x01,
	0x0a, 0x09, 0x45, 0x72, 0x72, 0x6f, 0x72, 0x43, 0x6f, 0x64, 0x65, 0x12, 0x61, 0x0a, 0x0a, 0x65,
	0x72, 0x72, 0x6f, 0x72, 0x5f, 0x63, 0x6f, 0x64, 0x65, 0x18, 0x01, 0x20, 0x01, 0x28, 0x0e, 0x32,
	0x42, 0x2e, 0x61, 0x65, 0x61, 0x2e, 0x66, 0x65, 0x74, 0x63, 0x68, 0x61, 0x69, 0x2e, 0x64, 0x65,
	0x66, 0x61, 0x75, 0x6c, 0x74, 0x2e, 0x76, 0x30, 0x5f, 0x31, 0x5f, 0x30, 0x2e, 0x44, 0x65, 0x66,
	0x61, 0x75, 0x6c, 0x74, 0x4d, 0x65, 0x73, 0x73, 0x61, 0x67, 0x65, 0x2e, 0x45, 0x72, 0x72, 0x6f,
	0x72, 0x43, 0x6f, 0x64, 0x65, 0x2e, 0x45, 0x72, 0x72, 0x6f, 0x72, 0x43, 0x6f, 0x64, 0x65, 0x45,
	0x6e, 0x75, 0x6d, 0x52, 0x09, 0x65, 0x72, 0x72, 0x6f, 0x72, 0x43, 0x6f, 0x64, 0x65, 0x22, 0x7f,
	0x0a, 0x0d, 0x45, 0x72, 0x72, 0x6f, 0x72, 0x43, 0x6f, 0x64, 0x65, 0x45, 0x6e, 0x75, 0x6d, 0x12,
	0x18, 0x0a, 0x14, 0x55, 0x4e, 0x53, 0x55, 0x50, 0x50, 0x4f, 0x52, 0x54, 0x45, 0x44, 0x5f, 0x50,
	0x52, 0x4f, 0x54, 0x4f, 0x43, 0x4f, 0x4c, 0x10, 0x00, 0x12, 0x12, 0x0a, 0x0e, 0x44, 0x45, 0x43,
	0x4f, 0x44, 0x49, 0x4e, 0x47, 0x5f, 0x45, 0x52, 0x52, 0x4f, 0x52, 0x10, 0x01, 0x12, 0x13, 0x0a,
	0x0f, 0x49, 0x4e, 0x56, 0x41, 0x4c, 0x49, 0x44, 0x5f, 0x4d, 0x45, 0x53, 0x53, 0x41, 0x47, 0x45,
	0x10, 0x02, 0x12, 0x15, 0x0a, 0x11, 0x55, 0x4e, 0x53, 0x55, 0x50, 0x50, 0x4f, 0x52, 0x54, 0x45,
	0x44, 0x5f, 0x53, 0x4b, 0x49, 0x4c, 0x4c, 0x10, 0x03, 0x12, 0x14, 0x0a, 0x10, 0x49, 0x4e, 0x56,
	0x41, 0x4c, 0x49, 0x44, 0x5f, 0x44, 0x49, 0x41, 0x4c, 0x4f, 0x47, 0x55, 0x45, 0x10, 0x04, 0x1a,
	0x2e, 0x0a, 0x12, 0x42, 0x79, 0x74, 0x65, 0x73, 0x5f, 0x50, 0x65, 0x72, 0x66, 0x6f, 0x72, 0x6d,
	0x61, 0x74, 0x69, 0x76, 0x65, 0x12, 0x18, 0x0a, 0x07, 0x63, 0x6f, 0x6e, 0x74, 0x65, 0x6e, 0x74,
	0x18, 0x01, 0x20, 0x01, 0x28, 0x0c, 0x52, 0x07, 0x63, 0x6f, 0x6e, 0x74, 0x65, 0x6e, 0x74, 0x1a,
	0xb1, 0x02, 0x0a, 0x12, 0x45, 0x72, 0x72, 0x6f, 0x72, 0x5f, 0x50, 0x65, 0x72, 0x66, 0x6f, 0x72,
	0x6d, 0x61, 0x74, 0x69, 0x76, 0x65, 0x12, 0x53, 0x0a, 0x0a, 0x65, 0x72, 0x72, 0x6f, 0x72, 0x5f,
	0x63, 0x6f, 0x64, 0x65, 0x18, 0x01, 0x20, 0x01, 0x28, 0x0b, 0x32, 0x34, 0x2e, 0x61, 0x65, 0x61,
	0x2e, 0x66, 0x65, 0x74, 0x63, 0x68, 0x61, 0x69, 0x2e, 0x64, 0x65, 0x66, 0x61, 0x75, 0x6c, 0x74,
	0x2e, 0x76, 0x30, 0x5f, 0x31, 0x5f, 0x30, 0x2e, 0x44, 0x65, 0x66, 0x61, 0x75, 0x6c, 0x74, 0x4d,
	0x65, 0x73, 0x73, 0x61, 0x67, 0x65, 0x2e, 0x45, 0x72, 0x72, 0x6f, 0x72, 0x43, 0x6f, 0x64, 0x65,
	0x52, 0x09, 0x65, 0x72, 0x72, 0x6f, 0x72, 0x43, 0x6f, 0x64, 0x65, 0x12, 0x1b, 0x0a, 0x09, 0x65,
	0x72, 0x72, 0x6f, 0x72, 0x5f, 0x6d, 0x73, 0x67, 0x18, 0x02, 0x20, 0x01, 0x28, 0x09, 0x52, 0x08,
	0x65, 0x72, 0x72, 0x6f, 0x72, 0x4d, 0x73, 0x67, 0x12, 0x6b, 0x0a, 0x0a, 0x65, 0x72, 0x72, 0x6f,
	0x72, 0x5f, 0x64, 0x61, 0x74, 0x61, 0x18, 0x03, 0x20, 0x03, 0x28, 0x0b, 0x32, 0x4c, 0x2e, 0x61,
	0x65, 0x61, 0x2e, 0x66, 0x65, 0x74, 0x63, 0x68, 0x61, 0x69, 0x2e, 0x64, 0x65, 0x66, 0x61, 0x75,
	0x6c, 0x74, 0x2e, 0x76, 0x30, 0x5f, 0x31, 0x5f, 0x30, 0x2e, 0x44, 0x65, 0x66, 0x61, 0x75, 0x6c,
	0x74, 0x4d, 0x65, 0x73, 0x73, 0x61, 0x67, 0x65, 0x2e, 0x45, 0x72, 0x72, 0x6f, 0x72, 0x5f, 0x50,
	0x65, 0x72, 0x66, 0x6f, 0x72, 0x6d, 0x61, 0x74, 0x69, 0x76, 0x65, 0x2e, 0x45, 0x72, 0x72, 0x6f,
	0x72, 0x44, 0x61, 0x74, 0x61, 0x45, 0x6e, 0x74, 0x72, 0x79, 0x52, 0x09, 0x65, 0x72, 0x72, 0x6f,
	0x72, 0x44, 0x61, 0x74, 0x61, 0x1a, 0x3c, 0x0a, 0x0e, 0x45, 0x72, 0x72, 0x6f, 0x72, 0x44, 0x61,
	0x74, 0x61, 0x45, 0x6e, 0x74, 0x72, 0x79, 0x12, 0x10, 0x0a, 0x03, 0x6b, 0x65, 0x79, 0x18, 0x01,
	0x20, 0x01, 0x28, 0x09, 0x52, 0x03, 0x6b, 0x65, 0x79, 0x12, 0x14, 0x0a, 0x05, 0x76, 0x61, 0x6c,
	0x75, 0x65, 0x18, 0x02, 0x20, 0x01, 0x28, 0x0c, 0x52, 0x05, 0x76, 0x61, 0x6c, 0x75, 0x65, 0x3a,
	0x02, 0x38, 0x01, 0x1a, 0x12, 0x0a, 0x10, 0x45, 0x6e, 0x64, 0x5f, 0x50, 0x65, 0x72, 0x66, 0x6f,
	0x72, 0x6d, 0x61, 0x74, 0x69, 0x76, 0x65, 0x42, 0x0e, 0x0a, 0x0c, 0x70, 0x65, 0x72, 0x66, 0x6f,
	0x72, 0x6d, 0x61, 0x74, 0x69, 0x76, 0x65, 0x62, 0x06, 0x70, 0x72, 0x6f, 0x74, 0x6f, 0x33,
}

var (
	file_default_proto_rawDescOnce sync.Once
	file_default_proto_rawDescData = file_default_proto_rawDesc
)

func file_default_proto_rawDescGZIP() []byte {
	file_default_proto_rawDescOnce.Do(func() {
		file_default_proto_rawDescData = protoimpl.X.CompressGZIP(file_default_proto_rawDescData)
	})
	return file_default_proto_rawDescData
}

var file_default_proto_enumTypes = make([]protoimpl.EnumInfo, 1)
var file_default_proto_msgTypes = make([]protoimpl.MessageInfo, 6)
var file_default_proto_goTypes = []interface{}{
	(DefaultMessage_ErrorCode_ErrorCodeEnum)(0), // 0: aea.fetchai.default.v0_1_0.DefaultMessage.ErrorCode.ErrorCodeEnum
	(*DefaultMessage)(nil),                      // 1: aea.fetchai.default.v0_1_0.DefaultMessage
	(*DefaultMessage_ErrorCode)(nil),            // 2: aea.fetchai.default.v0_1_0.DefaultMessage.ErrorCode
	(*DefaultMessage_Bytes_Performative)(nil),   // 3: aea.fetchai.default.v0_1_0.DefaultMessage.Bytes_Performative
	(*DefaultMessage_Error_Performative)(nil),   // 4: aea.fetchai.default.v0_1_0.DefaultMessage.Error_Performative
	(*DefaultMessage_End_Performative)(nil),     // 5: aea.fetchai.default.v0_1_0.DefaultMessage.End_Performative
	nil,                                         // 6: aea.fetchai.default.v0_1_0.DefaultMessage.Error_Performative.ErrorDataEntry
}
var file_default_proto_depIdxs = []int32{
	3, // 0: aea.fetchai.default.v0_1_0.DefaultMessage.bytes:type_name -> aea.fetchai.default.v0_1_0.DefaultMessage.Bytes_Performative
	5, // 1: aea.fetchai.default.v0_1_0.DefaultMessage.end:type_name -> aea.fetchai.default.v0_1_0.DefaultMessage.End_Performative
	4, // 2: aea.fetchai.default.v0_1_0.DefaultMessage.error:type_name -> aea.fetchai.default.v0_1_0.DefaultMessage.Error_Performative
	0, // 3: aea.fetchai.default.v0_1_0.DefaultMessage.ErrorCode.error_code:type_name -> aea.fetchai.default.v0_1_0.DefaultMessage.ErrorCode.ErrorCodeEnum
	2, // 4: aea.fetchai.default.v0_1_0.DefaultMessage.Error_Performative.error_code:type_name -> aea.fetchai.default.v0_1_0.DefaultMessage.ErrorCode
	6, // 5: aea.fetchai.default.v0_1_0.DefaultMessage.Error_Performative.error_data:type_name -> aea.fetchai.default.v0_1_0.DefaultMessage.Error_Performative.ErrorDataEntry
	6, // [6:6] is the sub-list for method output_type
	6, // [6:6] is the sub-list for method input_type
	6, // [6:6] is the sub-list for extension type_name
	6, // [6:6] is the sub-list for extension extendee
	0, // [0:6] is the sub-list for field type_name
}

func init() { file_default_proto_init() }
func file_default_proto_init() {
	if File_default_proto != nil {
		return
	}
	if !protoimpl.UnsafeEnabled {
		file_default_proto_msgTypes[0].Exporter = func(v interface{}, i int) interface{} {
			switch v := v.(*DefaultMessage); i {
			case 0:
				return &v.state
			case 1:
				return &v.sizeCache
			case 2:
				return &v.unknownFields
			default:
				return nil
			}
		}
		file_default_proto_msgTypes[1].Exporter = func(v interface{}, i int) interface{} {
			switch v := v.(*DefaultMessage_ErrorCode); i {
			case 0:
				return &v.state
			case 1:
				return &v.sizeCache
			case 2:
				return &v.unknownFields
			default:
				return nil
			}
		}
		file_default_proto_msgTypes[2].Exporter = func(v interface{}, i int) interface{} {
			switch v := v.(*DefaultMessage_Bytes_Performative); i {
			case 0:
				return &v.state
			case 1:
				return &v.sizeCache
			case 2:
				return &v.unknownFields
			default:
				return nil
			}
		}
		file_default_proto_msgTypes[3].Exporter = func(v interface{}, i int) interface{} {
			switch v := v.(*DefaultMessage_Error_Performative); i {
			case 0:
				return &v.state
			case 1:
				return &v.sizeCache
			case 2:
				return &v.unknownFields
			default:
				return nil
			}
		}
		file_default_proto_msgTypes[4].Exporter = func(v interface{}, i int) interface{} {
			switch v := v.(*DefaultMessage_End_Performative); i {
			case 0:
				return &v.state
			case 1:
				return &v.sizeCache
			case 2:
				return &v.unknownFields
			default:
				return nil
			}
		}
	}
	file_default_proto_msgTypes[0].OneofWrappers = []interface{}{
		(*DefaultMessage_Bytes)(nil),
		(*DefaultMessage_End)(nil),
		(*DefaultMessage_Error)(nil),
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: file_default_proto_rawDesc,
			NumEnums:      1,
			NumMessages:   6,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_default_proto_goTypes,
		DependencyIndexes: file_default_proto_depIdxs,
		EnumInfos:         file_default_proto_enumTypes,
		MessageInfos:      file_default_proto_msgTypes,
	}.Build()
	File_default_proto = out.File
	file_default_proto_rawDesc = nil
	file_default_proto_goTypes = nil
	file_default_proto_depIdxs = nil
}
